/**
far away - a free space shooter game
Copyright (C) 2005  Alexander Clausen, Dirk Rathlev, Stefan Toller

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include "SDL.h"
#include "SDL_image.h"
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>

#include "faraway.h"
#include "time.h"

#ifndef WIN32
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#endif

#ifndef DATA_DIR
#define	DATA_DIR	"/"
#endif

SDL_Surface *screen;

SDL_Surface *images[MAX_IMAGES];

faMain* g_pfaMain = NULL;

int faMain::game_init()
{
	// colorkey for transparency
	Uint32 colorkey;
	
	this->fTime = 0.0f;

	for(int i = 0; i<5; i++)
		faverage_fTime[i] = 0.0f;
	
#ifndef WIN32 /* on windows, we always use the local data dir, on everything else, we look for both */
	struct stat s;
	
	/* find data dir */
	if(0 != stat("data/objtypes", &s)) { /* no local data dir found */
		if(0 != stat(DATA_DIR, &s)) {
			perror("couldn't find data dir");
			return 6;
		}
		if(0 != chdir(DATA_DIR)) {
			perror("couldn't chdir into data dir");
			return 7;
		}
		if(0 != stat("data/objtypes", &s)) { /* no global data dir found */
			perror("stat failed");
			printf("no data found in %s\n", DATA_DIR);
			return 8;
		}
		printf("using global data dir\n");
	} else {
		printf("using local data dir\n");
	}
#endif
	
	g_pInput = new Input;
	g_pObjects = new Objects;
	g_pSound = new Sound;
	g_pUI = new UI;
	g_pCampaign = new Campaign;
	
	if (g_pInput->input_init())
	{
		printf("wasn't able  to initialize input subsystem\n");
		return 1;
	}
	
	if (g_pObjects->objects_init())
	{
		printf("wasn't able  to initialize objects subsystem\n");
		return 2;
	}
		
	if (Configs.bSoundEnabled) {
		if (g_pSound->sound_init())
		{
			printf("wasn't able  to initialize sound subsystem\n");
			return 3;
		}
	}
	
	if (g_pUI->UI_init())
	{
		printf("wasn't able  to initialize UI subsystem\n");
		return 4;
	}

	if (g_pCampaign->campaign_init())
	{
		printf("wasn't able  to initialize campaign subsystem\n");
		return 5;
	}
	
	unsigned int i=0;
	
	// load the image into a temporary surface first, then convert it into a faster format
	// and store it in the images[] array

	images[0] = IMG_Load("graphics/obj_asteroid.png");
	images[1] = IMG_Load("graphics/gras2.bmp");
	images[2] = IMG_Load("graphics/ship_player.png");
	images[3] = IMG_Load("graphics/weapon_laserI.png");
	images[4] = IMG_Load("graphics/pwrup_health.png");
	images[5] = IMG_Load("graphics/ship_enemy1.png");
	images[6] = IMG_Load("graphics/ship_enemy2.png");
	images[7] = IMG_Load("graphics/ship_frigate.png");
	images[8] = IMG_Load("graphics/weapon_plasma_laser.png");
	images[9] = IMG_Load("graphics/weapon_laserII.png"); /* old one */
	images[10] = IMG_Load("graphics/weapon_fast_laser.png");
	images[11] = IMG_Load("graphics/pwrup_laser1.png");
	images[12] = IMG_Load("graphics/pwrup_laser2.png");
	images[13] = IMG_Load("graphics/pwrup_laser3.png");
	images[14] = IMG_Load("graphics/pwrup_speed.png");
	images[15] = IMG_Load("graphics/pwrup_regenerate.png");
	// R=246 G=13 B=227
	printf("loaded\n");

	// check whether all images were succesful loaded
	// if yes, set colorkey and transform them into a faster format
	for (i=0; i<NUM_IMAGES; i++)
	{
		if (images[i] == NULL)
		{
			fprintf (stderr, "image %d not found\n",i, SDL_GetError ());
			return 6;
		}
		else
		{
			colorkey = SDL_MapRGB(images[i]->format, 246, 13, 227);
//			SDL_SetColorKey( images[i],SDL_SRCCOLORKEY|SDL_RLEACCEL|SDL_HWACCEL, colorkey );
			images[i] = SDL_DisplayFormatAlpha(images[i]);
		}
	}

	printf("images sucessful loaded\n");
	
		// initialize some game variables
	nScroll_x = 0;
	nScroll_y = 0;
	
	game_speed = 50.0f;
	printf("game speed: %f\n",game_speed);

	printf("initialized game\n");
	return 0;
}

void faMain::game_exit()
{
	int i;
	// free surfaces
	for (i=0; i<NUM_IMAGES; i++)
	{
		if (images[i])
		{
			SDL_FreeSurface(images[i]);
		}
	}
	
	g_pObjects->exit_objects();
	if (Configs.bSoundEnabled) {
		g_pSound->exit_sound();
	}
	g_pUI->exit_UI();
	
	//free(g_pInput);
	//free(g_pObjects);

	printf("game deinitialized\n");
	
	SDL_Quit();
}

void faMain::render(SDL_Surface *screen)
{
	SDL_Rect dest,src;
	unsigned int i, leftright = 0, size = g_pObjects->lObjects.size();
	
	// clear surface
	SDL_FillRect(screen,NULL,SDL_MapRGB(screen->format, 0, 0, 0));

	// walk through object-array and draw all existable objects
	for (i=0; i<size; i++)
	{
		if (g_pObjects->lObjects[i].ObjType)
		{
			// don't draw objects which aren't on the screen
			if ( (g_pObjects->lObjects[i].x+g_pObjects->lObjects[i].ObjType->w < 0) || (g_pObjects->lObjects[i].x > SCREEN_W) ) {
				continue;
			}
			leftright = 0; /* default orientation: left-to-right */
			
			if(g_pObjects->lObjects[i].ObjType->nType == OBJ_TYPE_SHOT)
			{
				/* shot flying from right to left? */
				if(g_pObjects->lObjects[i].speed_x < 0)
				{
					leftright = 1; /* right-to-left */
				}
			}
			src.x = g_pObjects->lObjects[i].nanim*g_pObjects->lObjects[i].ObjType->w;
			src.y = (leftright+g_pObjects->lObjects[i].nanimtype)*g_pObjects->lObjects[i].ObjType->h;
			src.w = g_pObjects->lObjects[i].ObjType->w;
			src.h = g_pObjects->lObjects[i].ObjType->h;

			dest.x = (int) g_pObjects->lObjects[i].x-nScroll_x;
			dest.y = (int) g_pObjects->lObjects[i].y-nScroll_y;
			dest.w = g_pObjects->lObjects[i].ObjType->w;
			dest.h = g_pObjects->lObjects[i].ObjType->h;
			if (SDL_BlitSurface(images[g_pObjects->lObjects[i].ObjType->nImage], &src, screen, &dest))
			{
				printf("error while drawing graphic %d\n",g_pObjects->lObjects[i].ObjType->nImage);
			}
		}
	}

	// render the user interface
	g_pUI->render_ui(screen);
	
	// change backbuffer and screen
	SDL_Flip (screen);
}

int faMain::game_loop()
{
	// fills the array with the frame_time, that we can calculate the mean value of the fps
	int add_time = 0; 

	while(GameState != GS_QUIT)
	{
		// Startzeitpunkt messen
		Time1 = SDL_GetTicks();


		g_pInput->eventloop();
		g_pInput->process_input();
		
		if (GameState == GS_START_MENU)
		{
			g_pUI->process_start_menu();
			g_pUI->render_start_menu(screen);
			SDL_Delay(10);
		}
		
		if (GameState == GS_GAME)
		{
			//printf("fTime: %f, fps: %f\n",fTime,(1/fTime));
			//printf("%f/%d\n",way_passed,level_length);
			if (nTargetType == TARGET_TIME)
			{
				// increase by passed way
				way_passed += fTime;
			}
			else if (nTargetType == TARGET_WAY)
			{
				// increase by time
				way_passed += game_speed*fTime;
			}
			if (way_passed > level_length)
			{
				g_pfaMain->GameState = GS_SPLASH;
				g_pfaMain->continue_at = GS_NEXT_LEVEL;
				g_pfaMain->splash = 1;
			}
			g_pObjects->process_objects();
			
			// GameState still GS_GAME?
			if (GameState == GS_GAME)
			{
				render(screen);
			}
		}
		
		if (GameState == GS_MENU)
		{
			g_pUI->process_ingame_menu();
			g_pUI->render_ingame_menu(screen);
			SDL_Delay(10);
		}
		
		if (GameState == GS_GAMEOVER) {
			GameState = GS_START_MENU;
		}
		
		if (GameState == GS_SPLASH) {
			g_pUI->show_splashscreen(screen,splash);
			SDL_Delay(10);
		}
		
		if (GameState == GS_NEXT_LEVEL) {
			g_pCampaign->next_level();
		}

		// save end-time of the frame and calculate the duration in seconds
		Time2 = SDL_GetTicks();

		float fTime2 = (float)(Time2 - Time1) / 1000.0f;
		faverage_fTime[add_time] = fTime2;
		if(add_time<4)
			add_time++;
		else
			add_time=0;

		fTime = 0.0f;
		for(int i=0; i<5;i++)
			fTime += faverage_fTime[i];

		fTime /= 5;

		// delay 10 ms to give other processes some cpu time
		SDL_Delay(1);
	}
	return 0;
}

bool faMain::get_bool_from_string(char* s) {
	if (0 == strcmp(s,"yes")) {
		return true;
	} else if (0 == strcmp(s,"no")) {
		return false;
	} else {
		printf("illegal value in %s\n",s);
		return false;
	}
}

int faMain::get_int_from_string(char* s) {
	return atoi(s);
}

int faMain::load_configs(const char* filename)
{
	FILE* file;
	char sName[20];
	char sValue[3];
	
	// fill Configs with some default values
	Configs.bDoubleBuffering = true;
	Configs.bHWSurface = true;
	Configs.bFullscreen = false;
	Configs.bSoundEnabled = false;
	Configs.bMusicEnabled = false;
	Configs.nSampleRate = 44100;
	Configs.nSoundBuffer = 2048;
	
	printf("open config file \"%s\" ...\n",filename);
	
	file = fopen(filename,"r");
	if (file == NULL)
	{
		return 1;
	}
	while(0 == feof(file)) {
		fscanf(file,"%s\t%s\n",sName,sValue);
		//printf("sName: %s sValue: %s\n",sName,sValue);
		if (0 == strcmp(sName,"doublebuffering")) {
			Configs.bDoubleBuffering = get_bool_from_string(sValue);
			if (Configs.bDoubleBuffering) {
				printf("config: hardware accelerated double buffering set\n");
			}
		} else if (0 == strcmp(sName,"hwsurface")) {
			Configs.bHWSurface = get_bool_from_string(sValue);
			if (Configs.bHWSurface) {
				printf("config: using hardware accelerated surface\n");
			}
		} else if (0 == strcmp(sName,"fullscreen")) {
			Configs.bFullscreen = get_bool_from_string(sValue);
			printf("config: fullscreen set\n");
		} else if (0 == strcmp(sName,"sound_on")) {
			Configs.bSoundEnabled = get_bool_from_string(sValue);
			if (Configs.bSoundEnabled) {
				printf("config: sound_on set\n");
				printf("WARNING: sound not supportet yet\n");
			}
		} else if (0 == strcmp(sName,"music_on")) {
			Configs.bMusicEnabled = get_bool_from_string(sValue);
			if (Configs.bMusicEnabled) {
				printf("config: music_on set\n");
				printf("WARNING: music not supportet yet\n");
			}
		} else if (0 == strcmp(sName,"samplerate")) {
			Configs.nSampleRate = get_int_from_string(sValue);
			printf("config: sample rate set to %d\n",Configs.nSampleRate);
		} else if (0 == strcmp(sName,"soundbuffer")) {
			Configs.nSoundBuffer = get_int_from_string(sValue);
			printf("config: sound buffer set to %d\n",Configs.nSoundBuffer);
		} 
	}
	
	return 0;
}

int main (int argc, char **argv)
{
	int error;
	
	// initialize main class
	g_pfaMain = new faMain;
	
	// load configs
	if (g_pfaMain->load_configs("config")) {
		printf("ERROR: config file couldn't be loaded\n");
	}

	// set game state
	g_pfaMain->GameState = GS_START_MENU;

	printf("GameState set to %d\n",g_pfaMain->GameState);

	// initialize SDL video system and check for errors
	if (SDL_Init (SDL_INIT_VIDEO|SDL_INIT_AUDIO) != 0)
	{
		printf ("ERROR: Wasn't able to initialize SDL: %s\n",
			SDL_GetError ());
		return EXIT_FAILURE;
	}
	// when the application gets killed (by user or by OS)
	// SDL_Quit() will be called
	atexit (SDL_Quit);

	// initialize video mode and store surface in screen
	Uint32 flags=0;
	
	if (g_pfaMain->Configs.bDoubleBuffering) {
		flags = flags|SDL_DOUBLEBUF;
	}
	if (g_pfaMain->Configs.bHWSurface) {
		flags = flags|SDL_HWSURFACE;
	}
	if (g_pfaMain->Configs.bFullscreen) {
		flags = flags|SDL_FULLSCREEN;
	}
	
	screen = SDL_SetVideoMode (SCREEN_W, SCREEN_H, 32, flags);
	
	if (screen == NULL)
	{
		printf ("ERROR: wasn't able to initialize video mode: "
				" %s\n", SDL_GetError ());
		return EXIT_FAILURE;
	}
	else
	{
		printf("VideoMode set\n");
	}

	// intialize game
	error = g_pfaMain->game_init();
	if (error) {
		printf("ERROR: game_init() failed with code %d\n",error);
		g_pfaMain->GameState = GS_QUIT;
	}
	else {
		printf("game initialized - GameState=%d\n",g_pfaMain->GameState);
	}

	// enter game loop
	g_pfaMain->game_loop();

	// quit game
	g_pfaMain->game_exit();

	printf ("Closed successful!\n");
	return EXIT_SUCCESS;
}
